@@ -18,9 +18,10 @@ MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
         use_bigints(sizeof(IV) < sizeof(int64_t)),
         check_required_fields(true),
         explicit_defaults(false),
-        encode_defaults(true),
+        encode_defaults(false),
         check_enum_values(true),
         generic_extension_methods(true),
+        implicit_maps(false),
         accessor_style(GetAndSet) {
     if (options_ref == NULL || !SvOK(options_ref))
         return;
@@ -40,6 +41,7 @@ MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
     BOOLEAN_OPTION(encode_defaults, encode_defaults);
     BOOLEAN_OPTION(check_enum_values, check_enum_values);
     BOOLEAN_OPTION(generic_extension_methods, generic_extension_methods);
+    BOOLEAN_OPTION(implicit_maps, implicit_maps);
 
     if (SV **value = hv_fetchs(options, "accessor_style", 0)) {
         const char *buf = SvPV_nolen(*value);
@@ -147,6 +149,29 @@ namespace {
 
         copy_and_bind(aTHX_ name, temp_name.c_str(), perl_package, mapperfield);
     }
+
+    bool is_map_entry(const MessageDef *message, bool check_implicit_map) {
+        if (message->mapentry())
+            return true;
+        if (!check_implicit_map)
+            return false;
+        const FieldDef *key = message->FindFieldByNumber(1);
+        const FieldDef *value = message->FindFieldByNumber(2);
+        const char *msg_name = message->name();
+        if (message->field_count() != 2 || message->oneof_count() != 0 ||
+                key == NULL || value == NULL)
+            return false;
+        if (strcmp(key->name(), "key") != 0 || strcmp(value->name(), "value") != 0)
+            return false;
+        size_t msg_name_len = strlen(msg_name);
+        if (msg_name_len < 6 || strcmp(msg_name + msg_name_len - 5, "Entry") != 0)
+            return false;
+        if (key->IsSequence() || value->IsSequence() ||
+                key->is_extension() || value->is_extension() ||
+                key->subdef() /* message or enum */)
+            return false;
+        return true;
+    }
 }
 
 void Dynamic::map_message(pTHX_ const string &message, const string &perl_package, const MappingOptions &options) {
@@ -261,9 +286,11 @@ void Dynamic::map_message(pTHX_ const Descriptor *descriptor, const string &perl
         croak("Message '%s' has already been mapped", descriptor->full_name().c_str());
     if (options.use_bigints)
         eval_pv("require Math::BigInt", 1);
-
     HV *stash = gv_stashpvn(perl_package.data(), perl_package.size(), GV_ADD);
     const MessageDef *message_def = def_builder.GetMessageDef(descriptor);
+    if (is_map_entry(message_def, options.implicit_maps))
+        // it's likely I will regret this const_cast<>
+        upb_msgdef_setmapentry(const_cast<MessageDef *>(message_def), true);
     Mapper *mapper = new Mapper(aTHX_ this, message_def, stash, options);
     const char *getter_prefix, *setter_prefix;
 
@@ -328,16 +355,26 @@ void Dynamic::map_message(pTHX_ const Descriptor *descriptor, const string &perl
         }
 
         copy_and_bind_field(aTHX_ "clear_field", "clear_", "", perl_package, mapperfield);
-        if (mapperfield->is_repeated()) {
+        if (mapperfield->is_map()) {
+            if (getter_prefix) {
+                copy_and_bind_field(aTHX_ "get_map_item", getter_prefix, "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "set_map_item", setter_prefix, "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "get_map", getter_prefix, "_map", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "set_map", setter_prefix, "_map", perl_package, mapperfield);
+            } else {
+                copy_and_bind_field(aTHX_ "get_or_set_map_item", "", "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "get_or_set_map", "", "_map", perl_package, mapperfield);
+            }
+        } else if (mapperfield->is_repeated()) {
             copy_and_bind_field(aTHX_ "add_item", "add_", "", perl_package, mapperfield);
             copy_and_bind_field(aTHX_ "list_size", "", "_size", perl_package, mapperfield);
             if (getter_prefix) {
-                copy_and_bind_field(aTHX_ "get_item", getter_prefix, "", perl_package, mapperfield);
-                copy_and_bind_field(aTHX_ "set_item", setter_prefix, "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "get_list_item", getter_prefix, "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "set_list_item", setter_prefix, "", perl_package, mapperfield);
                 copy_and_bind_field(aTHX_ "get_list", getter_prefix, "_list", perl_package, mapperfield);
                 copy_and_bind_field(aTHX_ "set_list", setter_prefix, "_list", perl_package, mapperfield);
             } else {
-                copy_and_bind_field(aTHX_ "get_or_set_item", "", "", perl_package, mapperfield);
+                copy_and_bind_field(aTHX_ "get_or_set_list_item", "", "", perl_package, mapperfield);
                 copy_and_bind_field(aTHX_ "get_or_set_list", "", "_list", perl_package, mapperfield);
             }
         } else {
