@@ -86,6 +86,22 @@ string Mapper::Field::full_name() const {
                field_def->name();
 }
 
+FieldDef::Type Mapper::Field::map_value_type() const {
+    const vector<Field> &map_fields = mapper->fields;
+
+    return map_fields[1].is_value ?
+        map_fields[1].field_def->type() :
+        map_fields[0].field_def->type();
+}
+
+const STD_TR1::unordered_set<int32_t> &Mapper::Field::map_enum_values() const {
+    const vector<Field> &map_fields = mapper->fields;
+
+    return map_fields[1].is_value ?
+        map_fields[1].enum_values :
+        map_fields[0].enum_values;
+}
+
 bool Mapper::DecoderHandlers::apply_defaults_and_check() {
     const vector<bool> &seen = seen_fields.back();
     const Mapper *mapper = mappers.back();
@@ -214,6 +230,40 @@ bool Mapper::DecoderHandlers::on_end_sequence(DecoderHandlers *cxt, const int *f
     return true;
 }
 
+Mapper::DecoderHandlers *Mapper::DecoderHandlers::on_start_map(DecoderHandlers *cxt, const int *field_index) {
+    THX_DECLARE_AND_GET;
+
+    cxt->mark_seen(field_index);
+    const Mapper *mapper = cxt->mappers.back();
+    SV *target = cxt->get_target(field_index);
+    HV *hv = NULL;
+
+    if (!SvROK(target)) {
+        hv = newHV();
+
+        SvUPGRADE(target, SVt_RV);
+        SvROK_on(target);
+        SvRV_set(target, (SV *) hv);
+    } else
+        hv = (HV *) SvRV(target);
+
+    cxt->mappers.push_back(mapper->fields[*field_index].mapper);
+    cxt->items.push_back((SV *) hv);
+    cxt->items.push_back(sv_newmortal());
+    cxt->items.push_back(NULL);
+
+    return cxt;
+}
+
+bool Mapper::DecoderHandlers::on_end_map(DecoderHandlers *cxt, const int *field_index) {
+    cxt->mappers.pop_back();
+    cxt->items.pop_back();
+    cxt->items.pop_back();
+    cxt->items.pop_back();
+
+    return true;
+}
+
 Mapper::DecoderHandlers *Mapper::DecoderHandlers::on_start_sub_message(DecoderHandlers *cxt, const int *field_index) {
     THX_DECLARE_AND_GET;
 
@@ -254,6 +304,22 @@ bool Mapper::DecoderHandlers::on_end_sub_message(DecoderHandlers *cxt, const int
     return true;
 }
 
+bool Mapper::DecoderHandlers::on_end_map_entry(DecoderHandlers *cxt, const int *field_index) {
+    size_t size = cxt->items.size();
+    HV *hash = (HV *) cxt->items[size - 3];
+    SV *key = (SV *) cxt->items[size - 2];
+    SV *value = (SV *) cxt->items[size - 1];
+
+    SvREFCNT_inc(value);
+    hv_store_ent(hash, key, value, 0);
+
+    if (SvPOK(key))
+        SvLEN_set(key, 0);
+    SvOK_off(key);
+
+    return true;
+}
+
 template<class T>
 bool Mapper::DecoderHandlers::on_nv(DecoderHandlers *cxt, const int *field_index, T val) {
     THX_DECLARE_AND_GET;
@@ -376,7 +442,15 @@ SV *Mapper::DecoderHandlers::get_target(const int *field_index) {
     const Field &field = mapper->fields[*field_index];
     SV *curr = items.back();
 
-    if (SvTYPE(curr) == SVt_PVAV) {
+    if (field.is_key) {
+        return items[items.size() - 2];
+    } else if (field.is_value) {
+        SV *sv = sv_newmortal();
+
+        items[items.size() - 1] = sv;
+
+        return sv;
+    } else if (SvTYPE(curr) == SVt_PVAV) {
         AV *av = (AV *) curr;
 
         return *av_fetch(av, av_top_index(av) + 1, 1);
@@ -414,16 +488,22 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
 
     env.ReportErrorsTo(&status);
     registry->ref();
+    pb_encoder = NULL;
+    pb_decoder = NULL;
+    json_encoder = NULL;
+    json_decoder = NULL;
     pb_encoder_handlers = Encoder::NewHandlers(message_def);
     json_encoder_handlers = Printer::NewHandlers(message_def, false /* XXX option */);
     decoder_handlers = Handlers::New(message_def);
     decode_explicit_defaults = options.explicit_defaults;
-    encode_defaults = options.encode_defaults;
+    encode_defaults = message_def->syntax() == UPB_SYNTAX_PROTO2 &&
+        options.encode_defaults;
     check_enum_values = options.check_enum_values;
 
     if (!decoder_handlers->SetEndMessageHandler(UpbMakeHandler(DecoderHandlers::on_end_message)))
         croak("Unable to set upb end message handler for %s", message_def->full_name());
 
+    bool map_entry = message_def->mapentry();
     bool has_required = false;
     for (MessageDef::const_field_iterator it = message_def->field_begin(), en = message_def->field_end(); it != en; ++it) {
         int index = fields.size();
@@ -445,10 +525,21 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
             field.name = newSVpvn_share(temp.data(), temp.size(), 0);
         }
         field.name_hash = SvSHARED_HASH(field.name);
-        field.has_default = false;
+        field.has_default = field.is_map = false;
         field.mapper = NULL;
         field.oneof_index = -1;
 
+        if (map_entry) {
+            field.is_key = field_def->number() == 1;
+            field.is_value = field_def->number() == 2;
+        }
+
+        if (field_def->label() == UPB_LABEL_REPEATED &&
+                field_def->type() == UPB_TYPE_MESSAGE &&
+                field_def->message_subdef()->mapentry()) {
+            field.is_map = true;
+        }
+
 #define GET_SELECTOR(KIND, TO) \
     ok = ok && pb_encoder_handlers->GetSelector(field_def, UPB_HANDLER_##KIND, &field.selector.TO)
 
@@ -464,14 +555,17 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
         case UPB_TYPE_FLOAT:
             GET_SELECTOR(FLOAT, primitive);
             SET_VALUE_HANDLER(float, on_nv<float>);
+            field.default_nv = field_def->default_float();
             break;
         case UPB_TYPE_DOUBLE:
             GET_SELECTOR(DOUBLE, primitive);
             SET_VALUE_HANDLER(double, on_nv<double>);
+            field.default_nv = field_def->default_double();
             break;
         case UPB_TYPE_BOOL:
             GET_SELECTOR(BOOL, primitive);
             SET_VALUE_HANDLER(bool, on_bool);
+            field.default_bool = field_def->default_bool();
             break;
         case UPB_TYPE_STRING:
         case UPB_TYPE_BYTES:
@@ -481,12 +575,17 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
             SET_HANDLER(StartString, on_start_string);
             SET_HANDLER(String, on_string);
             SET_HANDLER(EndString, on_end_string);
+            field.default_str = field_def->default_string(&field.default_str_len);
             break;
         case UPB_TYPE_MESSAGE:
             GET_SELECTOR(STARTSUBMSG, msg_start);
             GET_SELECTOR(ENDSUBMSG, msg_end);
-            SET_HANDLER(StartSubMessage, on_start_sub_message);
-            SET_HANDLER(EndSubMessage, on_end_sub_message);
+            if (field.is_map) {
+                SET_HANDLER(EndSubMessage, on_end_map_entry);
+            } else {
+                SET_HANDLER(StartSubMessage, on_start_sub_message);
+                SET_HANDLER(EndSubMessage, on_end_sub_message);
+            }
             has_default = false;
             break;
         case UPB_TYPE_ENUM: {
@@ -495,6 +594,7 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
                 SET_VALUE_HANDLER(int32_t, on_enum);
             else
                 SET_VALUE_HANDLER(int32_t, on_iv<int32_t>);
+            field.default_iv = field_def->default_int32();
 
             if (check_enum_values) {
                 const EnumDef *enumdef = field_def->enum_subdef();
@@ -507,10 +607,12 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
         case UPB_TYPE_INT32:
             GET_SELECTOR(INT32, primitive);
             SET_VALUE_HANDLER(int32_t, on_iv<int32_t>);
+            field.default_iv = field_def->default_int32();
             break;
         case UPB_TYPE_UINT32:
             GET_SELECTOR(UINT32, primitive);
             SET_VALUE_HANDLER(uint32_t, on_uv<uint32_t>);
+            field.default_uv = field_def->default_uint32();
             break;
         case UPB_TYPE_INT64:
             GET_SELECTOR(INT64, primitive);
@@ -518,6 +620,7 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
                 SET_VALUE_HANDLER(int64_t, on_bigiv);
             else
                 SET_VALUE_HANDLER(int64_t, on_iv<int64_t>);
+            field.default_i64 = field_def->default_int64();
             break;
         case UPB_TYPE_UINT64:
             GET_SELECTOR(UINT64, primitive);
@@ -525,6 +628,7 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
                 SET_VALUE_HANDLER(uint64_t, on_biguv);
             else
                 SET_VALUE_HANDLER(uint64_t, on_uv<uint64_t>);
+            field.default_u64 = field_def->default_uint64();
             break;
         default:
             croak("Unhandled field type %d for field '%s'", field_def->type(), field.full_name().c_str());
@@ -539,8 +643,13 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
         if (field_def->label() == UPB_LABEL_REPEATED) {
             GET_SELECTOR(STARTSEQ, seq_start);
             GET_SELECTOR(ENDSEQ, seq_end);
-            SET_HANDLER(StartSequence, on_start_sequence);
-            SET_HANDLER(EndSequence, on_end_sequence);
+            if (field.is_map) {
+                SET_HANDLER(StartSequence, on_start_map);
+                SET_HANDLER(EndSequence, on_end_map);
+            } else {
+                SET_HANDLER(StartSequence, on_start_sequence);
+                SET_HANDLER(EndSequence, on_end_sequence);
+            }
         }
 
 #undef GET_SELECTOR
@@ -576,6 +685,12 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
 }
 
 Mapper::~Mapper() {
+    if (json_decoder)
+        upb_env_free(&env, json_decoder);
+    upb_env_free(&env, json_encoder);
+    upb_env_free(&env, pb_decoder);
+    upb_env_free(&env, pb_encoder);
+
     for (vector<Field>::iterator it = fields.begin(), en = fields.end(); it != en; ++it)
         if (it->mapper)
             it->mapper->unref();
@@ -632,8 +747,9 @@ SV *Mapper::make_object(SV *data) const {
             HV * hv = newHV();
             sv_2mortal((SV *) hv);
             HV *orig = (HV *) SvRV(data);
-            I32 items = hv_iterinit(orig), keylen;
+            I32 keylen;
             char *key;
+            hv_iterinit(orig);
             while (SV *sv = hv_iternextsv(orig, &key, &keylen)) {
                 hv_store(hv, key, keylen, newSVsv(sv), 0);
             }
@@ -664,7 +780,7 @@ void Mapper::create_encoder_decoder() {
     json_decoder_method = ParserMethod::New(message_def);
     decoder_sink.Reset(decoder_handlers.get(), &decoder_callbacks);
     pb_decoder = upb::pb::Decoder::Create(&env, pb_decoder_method.get(), &decoder_sink);
-    json_decoder = upb::json::Parser::Create(&env, json_decoder_method.get(), &decoder_sink);
+    json_decoder = NULL;
     pb_encoder = upb::pb::Encoder::Create(&env, pb_encoder_handlers.get(), string_sink.input());
     json_encoder = upb::json::Printer::Create(&env, json_encoder_handlers.get(), string_sink.input());
 }
@@ -711,15 +827,23 @@ SV *Mapper::decode(const char *buffer, STRLEN bufsize) {
 }
 
 SV *Mapper::decode_json(const char *buffer, STRLEN bufsize) {
-    if (json_decoder == NULL)
+    if (pb_decoder == NULL)
         croak("It looks like resolve_references() was not called (and please use map() anyway)");
     status.Clear();
     decoder_callbacks.prepare(newHV());
+    // if an exception was thrown
+    if (json_decoder != NULL)
+        upb_env_free(&env, json_decoder);
+    // at the moment, the JSON parser can't be reused, and it's unclear whether
+    // this is a bug or a feature, so just create a new one every time
+    json_decoder = upb::json::Parser::Create(&env, json_decoder_method.get(), &decoder_sink);
 
     SV *result = NULL;
     if (BufferSource::PutBuffer(buffer, bufsize, json_decoder->input()))
         result = sv_bless(newRV_inc(decoder_callbacks.get_target()), stash);
     decoder_callbacks.clear();
+    upb_env_free(&env, json_decoder);
+    json_decoder = NULL;
 
     return result;
 }
@@ -788,6 +912,34 @@ namespace {
             return SvIV(src);
     }
 
+    IV key_iv(const char *key, I32 keylen) {
+        UV value;
+        int numtype = grok_number(key, keylen, &value);
+
+        if (numtype & IS_NUMBER_IN_UV) {
+            if (numtype & IS_NUMBER_NEG) {
+                if (value < (UV) IV_MIN)
+                    return - (IV) value;
+            } else {
+                if (value < (UV) IV_MAX)
+                    return (IV) value;
+            }
+        }
+        // XXX warn
+        return 0;
+    }
+
+    UV key_uv(const char *key, I32 keylen) {
+        UV value;
+        int numtype = grok_number(key, keylen, &value);
+
+        if (numtype & IS_NUMBER_IN_UV) {
+            return value;
+        }
+        // XXX warn
+        return 0;
+    }
+
     class SVGetter {
     public:
         SV *operator()(pTHX_ SV *src) const { return src; }
@@ -982,10 +1134,14 @@ bool Mapper::encode(Sink *sink, Status *status, SV *ref) const {
                 last_oneof = it->oneof_index;
         }
 
-        if (it->field_def->label() == UPB_LABEL_REPEATED)
+        if (it->is_map)
+            ok = ok && encode_from_perl_hash(sink, status, *it, HeVAL(he));
+        else if (it->field_def->label() == UPB_LABEL_REPEATED)
             ok = ok && encode_from_perl_array(sink, status, *it, HeVAL(he));
-        else
+        else if (encode_defaults || !it->has_default)
             ok = ok && encode(sink, status, *it, HeVAL(he));
+        else
+            ok = ok && encode_nodefaults(sink, status, *it, HeVAL(he));
     }
 
     if (!sink->EndMessage(status))
@@ -1053,6 +1209,171 @@ bool Mapper::encode(Sink *sink, Status *status, const Field &fd, SV *ref) const
     }
 }
 
+bool Mapper::encode_nodefaults(Sink *sink, Status *status, const Field &fd, SV *ref) const {
+    switch (fd.field_def->type()) {
+    case UPB_TYPE_FLOAT: {
+        NV value = SvNV(ref);
+        if (value == fd.default_nv)
+            return true;
+        return sink->PutFloat(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_DOUBLE: {
+        NV value = SvNV(ref);
+        if (value == fd.default_nv)
+            return true;
+        return sink->PutDouble(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_BOOL: {
+        bool value = SvNV(ref);
+        if (value == fd.default_bool)
+            return true;
+        return sink->PutBool(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_STRING:
+    case UPB_TYPE_BYTES: {
+        STRLEN len;
+        const char *str = fd.field_def->type() == UPB_TYPE_STRING ? SvPVutf8(ref, len) : SvPV(ref, len);
+        if (len == fd.default_str_len &&
+                (len == 0 || memcmp(str, fd.default_str, len) == 0))
+            return true;
+        Sink sub;
+        if (!sink->StartString(fd.selector.str_start, len, &sub))
+            return false;
+        sub.PutStringBuffer(fd.selector.str_cont, str, len, NULL);
+        return sink->EndString(fd.selector.str_end);
+    }
+    case UPB_TYPE_ENUM: {
+        IV value = SvIV(ref);
+        if (value == fd.default_iv)
+            return true;
+        if (check_enum_values &&
+                fd.enum_values.find(value) == fd.enum_values.end()) {
+            status->SetFormattedErrorMessage(
+                "Invalid enumeration value %d for field '%s'",
+                value,
+                fd.full_name().c_str()
+            );
+            return false;
+        }
+
+        return sink->PutInt32(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_INT32: {
+        IV value = SvIV(ref);
+        if (value == fd.default_iv)
+            return true;
+        return sink->PutInt32(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_UINT32: {
+        UV value = SvUV(ref);
+        if (value == fd.default_uv)
+            return true;
+        return sink->PutUInt32(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_INT64: {
+        int64_t value = sizeof(IV) >= sizeof(int64_t) ? SvIV(ref) : get_int64(aTHX_ ref);
+        if (value == fd.default_i64)
+            return true;
+        return sink->PutInt64(fd.selector.primitive, value);
+    }
+    case UPB_TYPE_UINT64: {
+        uint64_t value = sizeof(UV) >= sizeof(int64_t) ? SvUV(ref) : get_uint64(aTHX_ ref);
+        if (value == fd.default_u64)
+            return true;
+        return sink->PutUInt64(fd.selector.primitive, value);
+    }
+    default:
+        return false; // just in case
+    }
+}
+
+bool Mapper::encode_key(Sink *sink, Status *status, const Field &fd, const char *key, I32 keylen) const {
+    switch (fd.field_def->type()) {
+    case UPB_TYPE_BOOL: {
+        // follows what SvTRUE() does for strings
+        bool bval = keylen > 1 || keylen == 1 && key[0] != '0';
+        return sink->PutBool(fd.selector.primitive, bval);
+    }
+    case UPB_TYPE_STRING: {
+        Sink sub;
+        if (!sink->StartString(fd.selector.str_start, keylen, &sub))
+            return false;
+        sub.PutStringBuffer(fd.selector.str_cont, key, keylen, NULL);
+        return sink->EndString(fd.selector.str_end);
+    }
+    case UPB_TYPE_INT32:
+        return sink->PutInt32(fd.selector.primitive, key_iv(key, keylen));
+    case UPB_TYPE_UINT32:
+        return sink->PutUInt32(fd.selector.primitive, key_uv(key, keylen));
+    case UPB_TYPE_INT64:
+        return sink->PutInt64(fd.selector.primitive, key_iv(key, keylen));
+    case UPB_TYPE_UINT64:
+        return sink->PutInt64(fd.selector.primitive, key_uv(key, keylen));
+    default:
+        return false; // just in case
+    }
+}
+
+bool Mapper::encode_hash_kv(Sink *sink, Status *status, const char *key, STRLEN keylen, SV *value) const {
+    if (!sink->StartMessage())
+        return false;
+    if (fields[0].is_key) {
+        if (!encode_key(sink, status, fields[0], key, keylen))
+            return false;
+        if (!encode(sink, status, fields[1], value))
+            return false;
+    } else {
+        if (!encode_key(sink, status, fields[1], key, keylen))
+            return false;
+        if (!encode(sink, status, fields[0], value))
+            return false;
+    }
+    if (!sink->EndMessage(status))
+        return false;
+    return true;
+}
+
+bool Mapper::encode_from_perl_hash(Sink *sink, Status *status, const Field &fd, SV *ref) const {
+    if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
+        croak("Not an hash reference when encoding field '%s'", fd.full_name().c_str());
+    HV *hash = (HV *) SvRV(ref);
+    Sink repeated;
+
+    if (!sink->StartSequence(fd.selector.seq_start, &repeated))
+        return false;
+
+    hv_iterinit(hash);
+    while (HE *entry = hv_iternext(hash)) {
+        Sink key_value;
+        SV *value = HeVAL(entry);
+        const char *key;
+        STRLEN keylen;
+        bool needs_free = false;
+
+        if (HeKLEN(entry) == HEf_SVKEY) {
+            key = SvPVutf8(HeKEY_sv(entry), keylen);
+        } else {
+            if (HeKUTF8(entry)) {
+                key = HeKEY(entry);
+                keylen = HeKLEN(entry);
+            } else {
+                keylen = HeKLEN(entry);
+                key = (const char *) bytes_to_utf8((U8*) HeKEY(entry), &keylen);
+                SAVEFREEPV(key);
+            }
+        }
+
+        if (!repeated.StartSubMessage(fd.selector.msg_start, &key_value))
+            return false;
+        if (!fd.mapper->encode_hash_kv(&key_value, status, key, keylen, value))
+            return false;
+        if (!repeated.EndSubMessage(fd.selector.msg_end))
+            return false;
+    }
+
+    return sink->EndSequence(fd.selector.seq_end);
+}
+
 bool Mapper::encode_from_perl_array(Sink *sink, Status *status, const Field &fd, SV *ref) const {
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVAV)
         croak("Not an array reference when encoding field '%s'", fd.full_name().c_str());
@@ -1312,6 +1633,46 @@ AV *MapperField::get_write_array(HV *self) {
     }
 }
 
+SV *MapperField::get_read_hash_ref(HV *self) {
+    HE *ent = hv_fetch_ent(self, field->name, 0, field->name_hash);
+
+    if (!ent)
+        return NULL;
+
+    SV *ref = HeVAL(ent);
+    if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
+        croak("Value of field '%s' is not an hash reference", field->full_name().c_str());
+
+    return ref;
+}
+
+HV *MapperField::get_read_hash(HV *self) {
+    SV *ref = get_read_hash_ref(self);
+
+    return ref ? (HV *) SvRV(ref) : NULL;
+}
+
+HV *MapperField::get_write_hash(HV *self) {
+    HE *ent = hv_fetch_ent(self, field->name, 1, field->name_hash);
+    SV *ref = HeVAL(ent);
+
+    if (!SvOK(ref)) {
+        HV *hv = newHV();
+
+        SvUPGRADE(ref, SVt_RV);
+        SvROK_on(ref);
+        SvRV_set(ref, (SV *) hv);
+
+        return hv;
+    } else {
+        SV *ref = HeVAL(ent);
+        if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
+            croak("Value of field '%s' is not an hash reference", field->full_name().c_str());
+
+        return (HV *) SvRV(ref);
+    }
+}
+
 const char *MapperField::name() {
     return field->field_def->name();
 }
@@ -1324,6 +1685,10 @@ bool MapperField::is_extension() {
     return field->field_def->is_extension();
 }
 
+bool MapperField::is_map() {
+    return field->is_map;
+}
+
 bool MapperField::has_field(HV *self) {
     return hv_fetch_ent(self, field->name, 0, field->name_hash);
 }
@@ -1432,7 +1797,7 @@ void MapperField::set_scalar(HV *self, SV *value) {
 void MapperField::copy_value(SV *target, SV *value) {
     const FieldDef *field_def = field->field_def;
 
-    switch (field_def->type()) {
+    switch (field->is_map ? field->map_value_type() : field_def->type()) {
     case UPB_TYPE_FLOAT:
         sv_setnv(target, SvNV(value));
         break;
@@ -1464,8 +1829,11 @@ void MapperField::copy_value(SV *target, SV *value) {
         break;
     case UPB_TYPE_ENUM: {
         I32 i32 = SvIV(value);
-        if (field->enum_values.size() &&
-                field->enum_values.find(i32) == field->enum_values.end())
+        const STD_TR1::unordered_set<int32_t> &enum_values = field->is_map ?
+            field->map_enum_values() :
+            field->enum_values;
+        if (enum_values.size() &&
+                enum_values.find(i32) == field->enum_values.end())
             croak("Invalid value %d for enumeration field '%s'", i32, field->full_name().c_str());
         sv_setiv(target, i32);
     }
@@ -1522,6 +1890,22 @@ SV *MapperField::get_item(HV *self, int index, SV *target) {
     }
 }
 
+SV *MapperField::get_item(HV *self, SV *key, SV *target) {
+    HV *hash = get_read_hash(self);
+
+    if (!hash)
+        croak("Accessing unset map field '%s'", field->full_name().c_str());
+    if (HvTOTALKEYS(hash) == 0)
+        croak("Accessing empty map field '%s'", field->full_name().c_str());
+    HE *value = hv_fetch_ent(hash, key, 0, 0);
+
+    if (value) {
+        return HeVAL(value);
+    } else {
+        croak("Accessing non-existing key '%s' for field '%s'", SvPV_nolen(key), field->full_name().c_str());
+    }
+}
+
 void MapperField::set_item(HV *self, int index, SV *value) {
     AV *array = get_write_array(self);
     SV **target = av_fetch(array, index, 1);
@@ -1529,6 +1913,13 @@ void MapperField::set_item(HV *self, int index, SV *value) {
     copy_value(*target, value ? value : NULL);
 }
 
+void MapperField::set_item(HV *self, SV *key, SV *value) {
+    HV *hash = get_write_hash(self);
+    HE *target = hv_fetch_ent(hash, key, 1, 0);
+
+    copy_value(HeVAL(target), value ? value : NULL);
+}
+
 void MapperField::add_item(HV *self, SV *value) {
     AV *array = get_write_array(self);
     SV **target = av_fetch(array, av_top_index(array) + 1, 1);
@@ -1568,6 +1959,29 @@ void MapperField::set_list(HV *self, SV *ref) {
     SvREFCNT_inc(SvRV(field_ref));
 }
 
+SV *MapperField::get_map(HV *self) {
+    SV *hash_ref = get_read_hash_ref(self);
+
+    return hash_ref ? hash_ref : &PL_sv_undef;
+}
+
+void MapperField::set_map(HV *self, SV *ref) {
+    if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
+        croak("Value for field '%s' is not an hash reference", field->full_name().c_str());
+    SV *field_ref = get_write_field(self);
+
+    if (!SvOK(field_ref)) {
+        SvUPGRADE(field_ref, SVt_RV);
+        SvROK_on(field_ref);
+    } else {
+        if (!SvROK(field_ref))
+            croak("Value of field '%s' is not a reference", field->full_name().c_str());
+        SvREFCNT_dec(SvRV(field_ref));
+    }
+    SvRV_set(field_ref, SvRV(ref));
+    SvREFCNT_inc(SvRV(field_ref));
+}
+
 EnumMapper::EnumMapper(pTHX_ Dynamic *_registry, const upb::EnumDef *_enum_def) :
         registry(_registry),
         enum_def(_enum_def) {
