@@ -47,11 +47,28 @@ public:
         SV *name;
         U32 name_hash;
         bool has_default;
+        bool is_map;
+        bool is_key;
+        bool is_value;
         const Mapper *mapper; // for Message/Group fields
         STD_TR1::unordered_set<int32_t> enum_values;
         int oneof_index;
+        union {
+            struct {
+                size_t default_str_len;
+                const char *default_str;
+            };
+            bool default_bool;
+            IV default_iv;
+            UV default_uv;
+            NV default_nv;
+            int64_t default_i64;
+            uint64_t default_u64;
+        };
 
         std::string full_name() const;
+        upb::FieldDef::Type map_value_type() const;
+        const STD_TR1::unordered_set<int32_t> &map_enum_values() const;
     };
 
     struct DecoderHandlers {
@@ -75,8 +92,11 @@ public:
         static bool on_end_string(DecoderHandlers *cxt, const int *field_index);
         static DecoderHandlers *on_start_sequence(DecoderHandlers *cxt, const int *field_index);
         static bool on_end_sequence(DecoderHandlers *cxt, const int *field_index);
+        static DecoderHandlers *on_start_map(DecoderHandlers *cxt, const int *field_index);
+        static bool on_end_map(DecoderHandlers *cxt, const int *field_index);
         static DecoderHandlers *on_start_sub_message(DecoderHandlers *cxt, const int *field_index);
         static bool on_end_sub_message(DecoderHandlers *cxt, const int *field_index);
+        static bool on_end_map_entry(DecoderHandlers *cxt, const int *field_index);
 
         template<class T>
         static bool on_nv(DecoderHandlers *cxt, const int *field_index, T val);
@@ -126,7 +146,11 @@ public:
 private:
     bool encode(upb::Sink *sink, upb::Status *status, SV *ref) const;
     bool encode(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
+    bool encode_nodefaults(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
+    bool encode_key(upb::Sink *sink, upb::Status *status, const Field &fd, const char *key, I32 keylen) const;
+    bool encode_hash_kv(upb::Sink *sink, upb::Status *status, const char *key, STRLEN keylen, SV *value) const;
     bool encode_from_perl_array(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
+    bool encode_from_perl_hash(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
     bool encode_from_message_array(upb::Sink *sink, upb::Status *status, const Mapper::Field &fd, AV *source) const;
 
     template<class G, class S>
@@ -175,6 +199,7 @@ public:
     const char *name();
     bool is_repeated();
     bool is_extension();
+    bool is_map();
 
     // presence
     bool has_field(HV *self);
@@ -192,6 +217,12 @@ public:
     SV *get_list(HV *self);
     void set_list(HV *self, SV *ref);
 
+    // map
+    SV *get_item(HV *self, SV *key, SV *target);
+    void set_item(HV *self, SV *key, SV *value);
+    SV *get_map(HV *self);
+    void set_map(HV *self, SV *ref);
+
     static MapperField *find_extension(pTHX_ CV *cv, SV *extension);
     static MapperField *find_scalar_extension(pTHX_ CV *cv, SV *extension);
     static MapperField *find_repeated_extension(pTHX_ CV *cv, SV *extension);
@@ -203,6 +234,9 @@ private:
     SV *get_read_array_ref(HV *self);
     AV *get_read_array(HV *self);
     AV *get_write_array(HV *self);
+    SV *get_read_hash_ref(HV *self);
+    HV *get_read_hash(HV *self);
+    HV *get_write_hash(HV *self);
     void copy_default(SV *target);
     void copy_value(SV *target, SV *value);
     void clear_oneof(HV *self);
